"""Demonstrates low and high level pixel data access with type BaseBitmap, including alpha
channel data.

To take full effect of this script you must open the file from within the provided directory, so
that the script can find the example files next to it.

The script demonstrates what I dubbed here high- and low-level BaseBitmap access, using 
BaseBitmap.GetPixelDirect (high) and BaseBitmap.GetPixelCnt (low). It is not advisable to use
low-level access, as it (slightly) looses out in performance on high-level access in most cases and
is much more complicated to write and handle.

The reason is to get the low level access somewhat performant, we must read data line-by-line
and deal manually with the pixel formats. When data is read pixel-by-pixel with low-level access, 
the code will be much cleaner but also MUCH slower than the built in high-level and also pixel by
pixel access methods. This is because Python's struct library is very slow.

Note:
    * This code example has turned out much more technical and dense than I am usually comfortable 
    with but that is unfortunately unavoidable. When you are on S22+, just look at 
    GetBitmapDataHighLevel() it is very straight forward, is able to deal with any pixel format,
    and in most cases also the most performant solution.
    * BaseBitmap is a bit messy at the moment, the reason is that it is only and adapter these days
    for the underlying Image API. In C++ one can get the underlying Image API data with
    BaseBitmap::GetImageRef() which makes things less finicky. In Python this is currently not
    possible as the Image API has not been wrapped there.
"""

import c4d

import itertools
import functools
import os
import struct
import time
import typing

# Maps render data bit depth symbols to their respective number of total bits per pixel.
FORMATDEPTH_TO_BITS: dict[int, int] = {
    c4d.RDATA_FORMATDEPTH_8: 8 * 3,
    c4d.RDATA_FORMATDEPTH_16: 16 * 3,
    c4d.RDATA_FORMATDEPTH_32: 32 * 3,
}

# Maps Image API byte sizes per pixel to the respective color modes.
COLOR_BYTES_TO_MODES: dict[int, int] = {
    c4d.COLORBYTES_RGB: c4d.COLORMODE_RGB,
    c4d.COLORBYTES_ARGB: c4d.COLORMODE_ARGB,
    c4d.COLORBYTES_RGBw: c4d.COLORMODE_RGBw,
    c4d.COLORBYTES_ARGBw: c4d.COLORMODE_ARGBw,
    c4d.COLORBYTES_RGBf: c4d.COLORMODE_RGBf,
    c4d.COLORBYTES_ARGBf: c4d.COLORMODE_ARGBf,
}

# Translates relevant pixel formats to packing strings used by Python's #struct.
COLOR_BYTES_TO_PACK_STRING: dict = {
    # COLORBYTES_RGB is an alias for maxon::PIX which is an alias for unsigned char.
    c4d.COLORBYTES_RGB: "B",
    c4d.COLORBYTES_ARGB: "B",
    # COLORBYTES_RGBw is an alias for maxon::PIX_W which is an alias for unsigned short/Int16.
    c4d.COLORBYTES_RGBw: "H",
    c4d.COLORBYTES_ARGBw: "H",
    # COLORBYTES_RGBw is an alias for maxon::PIX_F which is an alias for a single/32-bit float.
    c4d.COLORBYTES_RGBf: "f",
    c4d.COLORBYTES_ARGBf: "f",
}

# The active document.
doc: c4d.documents.BaseDocument
# Type alias for bitmap data stored as a dictionary of RGBA values over coordinate keys.
BitmapData: typing.Type = dict[tuple[int, int], c4d.Vector4d]

def TimeIt(func):
    """Provides a makeshift decorator for timing functions executions.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        """Wraps and measures the execution time of the wrapped function.
        """
        t0: int = time.perf_counter()
        result: typing.Any = func(*args, **kwargs)
        print(f"Executing {func.__name__}() took {(time.perf_counter() - t0):.4f} seconds.")
        return result
    return wrapper

def RenderImage(doc: c4d.documents.BaseDocument, size: tuple[int]) -> c4d.bitmaps.BaseBitmap:
    """Renders the active document into a BaseBitmap with the given #size.

    Does not ensure that the image/rendering is 32bits per channel.
    """
    data: c4d.BaseContainer = doc.GetActiveRenderData().GetData()
    data[c4d.RDATA_XRES] = size[0]
    data[c4d.RDATA_YRES] = size[1]

    bitFormat: int = data[c4d.RDATA_FORMATDEPTH]
    bmp: c4d.bitmaps.BaseBitmap = c4d.bitmaps.BaseBitmap()
    bmp.Init(size[0], size[1], FORMATDEPTH_TO_BITS[bitFormat])
    res: int = c4d.documents.RenderDocument(doc, data, bmp, c4d.RENDERFLAGS_EXTERNAL)
    if res != c4d.RENDERRESULT_OK:
        raise RuntimeError(f"Could not render {doc}.")

    return bmp

def LoadImages(path: str, format:str = ".psd") -> list[tuple[str, c4d.bitmaps.BaseBitmap]]:
    """Loads the images in #path with the given #format as BaseBitmap instances.
    """
    if not isinstance(path, str) or not os.path.exists(path):
        raise IOError()

    result: list[c4d.bitmaps.BaseBitmap] = []
    for fileName in os.listdir(path):
        file, ext = os.path.splitext(fileName)
        if not os.path.isfile(fileName) or ext.lower() != format:
            continue

        filePath: str = os.path.join(path, fileName)
        bmp: c4d.bitmaps.BaseBitmap = c4d.bitmaps.BaseBitmap()
        res, _ = bmp.InitWith(filePath)
        if res != c4d.IMAGERESULT_OK:
            raise IOError()

        result.append((fileName, bmp))

    return result

@TimeIt
def GetBitmapDataHighLevel(bmp: c4d.bitmaps.BaseBitmap) -> BitmapData:
    """Returns the bitmap data of #bmp via the high level methods BaseBitmap.GetPixelDirect and
    GetAlphaPixel.

    GetPixelDirect supports bit depths higher than 8-bit although its value lie in the typical 8-bit
    [0, 255] interval. The values are however floating point and not integer.
    """
    size: list[int] = bmp.GetSize()
    alphaChannel: typing.Optional[c4d.bitmaps.BaseBitmap] = bmp.GetInternalChannel()
    data: BitmapData = {}
    f: float = 1. / 255.
    for x, y in itertools.product(range(size[0]), range(size[1])):
        rgb: c4d.Vector = bmp.GetPixelDirect(x, y)
        a: float = bmp.GetAlphaPixel(alphaChannel, x, y) * f if alphaChannel else 1.
        data[(x, y)] = c4d.Vector4d(rgb.x * f, rgb.y * f, rgb.z * f, a)
    return data

@TimeIt
def GetBitmapDataLowLevel(bmp: c4d.bitmaps.BaseBitmap) -> BitmapData:
    """Returns the bitmap data of #bmp via direct memory access.

    In C++, such low level direct memory access is much faster, but as usually the case such
    direct and raw memory access does not translate well to Python. Even in its most optimized form,
    it will be slower than high level access, it is therefore recommended to use GetPixelDirect and
    GetAlphaPixel when possible. The major culprit is the `struct` library of Python which is very
    slow.

    Note:
        We could also technically ignore the source format here and always pack everything into
        32-bits as #dstmode of #GetPixelCnt denotes the destination format and not the source format.
        I instead showed here how to deal with each format in its native form.
    """
    width, height = bmp.GetSize()
    data: BitmapData = {}

    # Determine if the bitmap has an internal alpha channel or not. Note that other than in the
    # Image API, a BaseBitmap returns many values as if it had three channels, even when there is
    # an alpha channel as the fourth channel per pixel.
    alphaChannel: typing.Optional[c4d.bitmaps.BaseBitmap] = bmp.GetInternalChannel()
    chanelCount: int = 4 if alphaChannel else 3

    # The bits and bytes per pixel and the size of a line buffer.
    bitsPerPixel: int  = bmp.GetBt()
    bytesPerPixel: int = bitsPerPixel // 8
    bufferSize: int = bmp.GetBpz() # equal to bytesPerPixel *  width

    # E.g. for 32 bits per channel and a width of 128:
    # bitsPerPixel = 96, bytesPerPixel = 12, bytesPerPixel *  width = 1536 , bufferSize = 1536

    # As described above, the bitmap always acts as if it has three channels when it comes to
    # its #bitsPerPixel, and while we MUST traverse the data with a stride that includes the
    # alpha channel, there is no meaningful data to be found there.
    colorBytes: int = int((bytesPerPixel / 3) * chanelCount)
    colorMode: int = COLOR_BYTES_TO_MODES[colorBytes]
    bufferSize = colorBytes * width
    rgbBuffer: bytearray = bytearray(bufferSize)
    # Because we we are unable to read the alpha channel implicitly, we require a second buffer.
    if alphaChannel:
        alphaBuffer: bytearray = bytearray(bufferSize)

    # Pick the correct unpacking format corresponding to the maxon:PIX_? format referenced by the
    # given #colorBytes. See #COLOR_BYTES_TO_PACK_STRING at the top of the file.
    formatString: str = COLOR_BYTES_TO_PACK_STRING[colorBytes] * (chanelCount * width)

    # Conversion factors for converting 8- and 16-bit values to a floating point representation.
    f8: float = 1. / 255.
    f16: float = 1. / 65535.

    def GetLineData(data: c4d.bitmaps.BaseBitmap, y: int, buffer: bytearray) -> list[float]:
        """Abstracts reading one line of pixels into a floating point format independent of the
        source format.
        """
        # Read the line #y into the line buffer.
        data.GetPixelCnt(0, y, width, buffer, colorBytes, colorMode, c4d.PIXELCNT_NONE)
        # Unpack the line into its native form, Uchar (8bit), Ushort (16bit), or single (32bit).
        result: list[typing.Any] = struct.unpack(formatString, buffer)

        # Doing this is optional, but I convert here the 8 bit and 16 bit integer data to a floating
        # point format, I do this primarily because that is what is being done by the high level 
        # access methods.
        if colorBytes in (c4d.COLORBYTES_RGB, c4d.COLORBYTES_ARGB):
            result = [n * f8 for n in result]
        if colorBytes in (c4d.COLORBYTES_RGBw, c4d.COLORBYTES_ARGBw):
            result = [n * f16 for n in result]

        return result

    # Read and unpack data line by line. When we access data pixel by pixel instead, access will be
    # very slow, the culprit is here primarily #struct which is notorious for being very slow. This 
    # only applies to Python, in C++ low level access will be much faster than high level access.
    for y in range(height):
        # Unpack a line of RGB data and optionally alpha data when necessary.
        rgbRaw: list[float] = GetLineData(bmp, y, rgbBuffer)
        if alphaChannel:
            alphaRaw: list[float] = GetLineData(alphaChannel, y, alphaBuffer)

        # Convert one line into RGBA vectors stored over their pixel coordinate.
        for x, offset in enumerate(range(0, len(rgbRaw), chanelCount)):
            data[(x, y)] = c4d.Vector4d(rgbRaw[offset+1 if alphaChannel else offset],
                                        rgbRaw[offset+2 if alphaChannel else offset+1],
                                        rgbRaw[offset+3 if alphaChannel else offset+2],
                                        alphaRaw[offset+1] if alphaChannel else 1.)

    return data

def main() -> None:
    """Runs the example.
    """
    # Clear the console and get the input images. The images loaded by LoadImages() are expected
    # to lie next to this script.
    c4d.CallCommand(13957)
    imageData: list[tuple[str, c4d.bitmaps.BaseBitmap]] = (
        [("rendering", RenderImage(doc, (128, 128)))] + LoadImages(os.path.dirname(__file__)))

    # Iterate over the images, access their data low- and high-level data, and print out timings
    # and results. High level access wins print much in every case, and low level access is much
    # more cumbersome, I would avoid dealing with it when possible.
    for label, bmp in imageData:
        print("\n" + ("-" * 80))
        print(f"{label} - {bmp.GetSize() = }, {bmp.GetBt() = }, {bmp.GetBpz() = }.")
        highLevelData: BitmapData = GetBitmapDataHighLevel(bmp)
        lowLevelData: BitmapData = GetBitmapDataLowLevel(bmp)

        for key in list(highLevelData.keys())[0:10]:
                print (f"{key} high: {highLevelData[key]}\n"
                    f"{' ' * (len(str(key)) + 1)}low : {lowLevelData[key]}")

if __name__ == "__main__":
    main()